// react:
import { 
// hooks:
useRef, useEffect, } from 'react';
// cssfn:
import { 
// writes css in javascript:
rule, states, style, vars, cssVars, } from '@cssfn/core'; // writes css in javascript
// reusable-ui utilities:
import { 
// hooks:
useTriggerRender, useEvent, useMountedFlag, useScheduleTriggerEvent, } from '@reusable-ui/hooks'; // react helper hooks
// reusable-ui features:
import { 
// hooks:
usesAnimation, } from '@reusable-ui/animation'; // animation stuff of UI
const [excitableVars] = cssVars({ prefix: 'ex', minify: false }); // shared variables: ensures the server-side & client-side have the same generated css variable names
{
    const { animationRegistry: { registerFilter, registerAnim } } = usesAnimation();
    registerFilter(excitableVars.filter);
    registerAnim(excitableVars.anim);
}
// parent is     `.excited` -or-  current is     `.excited`:
export const ifExcited = (styles) => rule(':is(.excited&, &.excited)', styles); // specificityWeight = 1 + (parent's specificityWeight)
// parent is not `.excited` -and- current is not `.excited`:
export const ifNotExcited = (styles) => rule(':where(&):not(:is(.excited&, &.excited))', styles); // specificityWeight = 1 + (parent's specificityWeight)
/**
 * Adds a capability of UI to highlight itself to attract user's attention.
 * @param config  A configuration of `excitableRule`.
 * @returns An `ExcitableStuff` represents an excitable state.
 */
export const usesExcitable = (config) => {
    return {
        excitableRule: () => style({
            // animation states:
            ...states([
                ifExcited({
                    ...vars({
                        [excitableVars.filter]: config?.filterExcite,
                        [excitableVars.anim]: config?.animExcite,
                    }),
                }),
            ]),
        }),
        excitableVars,
    };
};
export const useExcitable = (props) => {
    // fn props:
    const propExcited = props.excited ?? false;
    // fn states:
    /*
     * state is excited/unexcited based on [controllable excited]
     * [uncontrollable excited] is not supported
     */
    const excitedFn = propExcited /*controllable*/ ? 2 /* ExcitableState.Excited */ : 0 /* ExcitableState.Unexcited */;
    // states:
    const isMounted = useMountedFlag();
    // local storages without causing to (re)render, we need to manual control the (re)render event:
    /**
     * `true`  => has excited
     * `false` => has normal
     * `null`  => need to restart
     */
    const stateRef = useRef(excitedFn);
    // manually controls the (re)render event:
    const [triggerRender] = useTriggerRender();
    // update state:
    if (stateRef.current !== 1 /* ExcitableState.Continue */) {
        if (stateRef.current !== excitedFn) { // change detected => apply the change & start animating
            stateRef.current = excitedFn; // remember the last change
            triggerRender(); // need to apply the animation
        } // if
    } // if
    // dom effects:
    useEffect(() => {
        // conditions:
        if (stateRef.current !== 1 /* ExcitableState.Continue */)
            return; // only process `Continue` (need to restart the animation)
        // setups:
        requestAnimationFrame(() => {
            // conditions:
            if (!isMounted.current)
                return;
            // actions:
            stateRef.current = 2 /* ExcitableState.Excited */; // restart
            triggerRender(); // need to restart the animation
        });
    }); // no dependency-list, runs every re-render
    // handlers:
    const handleAnimationStart = useEvent((event) => {
        // conditions:
        if (event.target !== event.currentTarget)
            return; // ignores bubbling
        if (!/((^|[^a-z])(excite)|([a-z])(Excite))(?![a-z])/.test(event.animationName))
            return; // ignores animation other than excite[Foo] or booExcite[Foo]
        // mark the animation has (re)started -- if was marked as `ExcitableState.Continue`:
        stateRef.current = 2 /* ExcitableState.Excited */;
    });
    const handleAnimationEnd = useEvent((event) => {
        // conditions:
        if (event.target !== event.currentTarget)
            return; // ignores bubbling
        if (!/((^|[^a-z])(excite)|([a-z])(Excite))(?![a-z])/.test(event.animationName))
            return; // ignores animation other than excite[Foo] or booExcite[Foo]
        // mark the animation has stopped:
        if (!propExcited) {
            stateRef.current = 0 /* ExcitableState.Unexcited */; // mark the animation has stopped
        }
        else {
            stateRef.current = 1 /* ExcitableState.Continue */; // mark the animation has stopped but need to restart
            triggerRender(); // need to restart the animation
        } // if
    });
    const handleAnimationCancel = handleAnimationEnd;
    // api:
    return {
        excited: propExcited,
        state: stateRef.current,
        class: (stateRef.current === 2 /* ExcitableState.Excited */) ? 'excited' : null,
        handleAnimationStart,
        handleAnimationEnd,
        handleAnimationCancel,
    };
};
export const useControllableExcitable = (props, excitableApi) => {
    // states:
    const { state } = excitableApi;
    // events:
    const scheduleTriggerEvent = useScheduleTriggerEvent();
    if (state === 1 /* ExcitableState.Continue */) {
        const { onExcitedChange, } = props;
        // request to stop:
        if (onExcitedChange)
            scheduleTriggerEvent(() => {
                // fire `onExcitedChange` react event:
                onExcitedChange?.({ excited: false });
            });
    } // if
};
//#endregion excitable
